/*
 * Saviola DDPD Mini control / PIC12F1840 @ 2023
 * C Code for MPLAB XC8 compiler
 * Internal oscillator used @ 32MHz
 * This is a free software with NO WARRANTY
 */

// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection->INTOSC oscillator: I/O function on CLKIN pin
#pragma config WDTE = OFF    // Watchdog Timer Enable->WDT disabled
#pragma config PWRTE = OFF    // Power-up Timer Enable->PWRT disabled
#pragma config MCLRE = OFF    // MCLR Pin Function Select->MCLR/VPP pin function is digital input
#pragma config CP = ON    // Flash Program Memory Code Protection->Program memory code protection is enabled
#pragma config CPD = ON    // Data Memory Code Protection->Data memory code protection is enabled
#pragma config BOREN = ON    // Brown-out Reset Enable->Brown-out Reset enabled
#pragma config CLKOUTEN = OFF    // Clock Out Enable->CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin
#pragma config IESO = ON    // Internal/External Switchover->Internal/External Switchover mode is enabled
#pragma config FCMEN = ON    // Fail-Safe Clock Monitor Enable->Fail-Safe Clock Monitor is enabled

// CONFIG2
#pragma config WRT = OFF    // Flash Memory Self-Write Protection->Write protection off
#pragma config PLLEN = OFF    // PLL Enable->4x PLL disabled
#pragma config STVREN = ON    // Stack Overflow/Underflow Reset Enable->Stack Overflow or Underflow will cause a Reset
#pragma config BORV = LO    // Brown-out Reset Voltage Selection->Brown-out Reset Voltage (Vbor), low trip point selected.
#pragma config LVP = OFF    // Low-Voltage Programming Enable->High-voltage on MCLR/VPP must be used for programming

#include <xc.h>
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdio.h>

#pragma warning disable 520

#define _XTAL_FREQ 32000000

#define INPUT   1
#define OUTPUT  0

#define HIGH    1
#define LOW     0

#define ANALOG      1
#define DIGITAL     0

#define PULL_UP_ENABLED      1
#define PULL_UP_DISABLED     0

// get/set IO_RA0 aliases
#define IO_RA0_TRIS                 TRISAbits.TRISA0
#define IO_RA0_LAT                  LATAbits.LATA0
#define IO_RA0_PORT                 PORTAbits.RA0
#define IO_RA0_WPU                  WPUAbits.WPUA0
#define IO_RA0_ANS                  ANSELAbits.ANSA0
#define IO_RA0_SetHigh()            do { LATAbits.LATA0 = 1; } while(0)
#define IO_RA0_SetLow()             do { LATAbits.LATA0 = 0; } while(0)
#define IO_RA0_Toggle()             do { LATAbits.LATA0 = ~LATAbits.LATA0; } while(0)
#define IO_RA0_GetValue()           PORTAbits.RA0
#define IO_RA0_SetDigitalInput()    do { TRISAbits.TRISA0 = 1; } while(0)
#define IO_RA0_SetDigitalOutput()   do { TRISAbits.TRISA0 = 0; } while(0)
#define IO_RA0_SetPullup()          do { WPUAbits.WPUA0 = 1; } while(0)
#define IO_RA0_ResetPullup()        do { WPUAbits.WPUA0 = 0; } while(0)
#define IO_RA0_SetAnalogMode()      do { ANSELAbits.ANSA0 = 1; } while(0)
#define IO_RA0_SetDigitalMode()     do { ANSELAbits.ANSA0 = 0; } while(0)

// get/set SCL aliases
#define SCL_TRIS                 TRISAbits.TRISA1
#define SCL_LAT                  LATAbits.LATA1
#define SCL_PORT                 PORTAbits.RA1
#define SCL_WPU                  WPUAbits.WPUA1
#define SCL_ANS                  ANSELAbits.ANSA1
#define SCL_SetHigh()            do { LATAbits.LATA1 = 1; } while(0)
#define SCL_SetLow()             do { LATAbits.LATA1 = 0; } while(0)
#define SCL_Toggle()             do { LATAbits.LATA1 = ~LATAbits.LATA1; } while(0)
#define SCL_GetValue()           PORTAbits.RA1
#define SCL_SetDigitalInput()    do { TRISAbits.TRISA1 = 1; } while(0)
#define SCL_SetDigitalOutput()   do { TRISAbits.TRISA1 = 0; } while(0)
#define SCL_SetPullup()          do { WPUAbits.WPUA1 = 1; } while(0)
#define SCL_ResetPullup()        do { WPUAbits.WPUA1 = 0; } while(0)
#define SCL_SetAnalogMode()      do { ANSELAbits.ANSA1 = 1; } while(0)
#define SCL_SetDigitalMode()     do { ANSELAbits.ANSA1 = 0; } while(0)

// get/set SDA aliases
#define SDA_TRIS                 TRISAbits.TRISA2
#define SDA_LAT                  LATAbits.LATA2
#define SDA_PORT                 PORTAbits.RA2
#define SDA_WPU                  WPUAbits.WPUA2
#define SDA_ANS                  ANSELAbits.ANSA2
#define SDA_SetHigh()            do { LATAbits.LATA2 = 1; } while(0)
#define SDA_SetLow()             do { LATAbits.LATA2 = 0; } while(0)
#define SDA_Toggle()             do { LATAbits.LATA2 = ~LATAbits.LATA2; } while(0)
#define SDA_GetValue()           PORTAbits.RA2
#define SDA_SetDigitalInput()    do { TRISAbits.TRISA2 = 1; } while(0)
#define SDA_SetDigitalOutput()   do { TRISAbits.TRISA2 = 0; } while(0)
#define SDA_SetPullup()          do { WPUAbits.WPUA2 = 1; } while(0)
#define SDA_ResetPullup()        do { WPUAbits.WPUA2 = 0; } while(0)
#define SDA_SetAnalogMode()      do { ANSELAbits.ANSA2 = 1; } while(0)
#define SDA_SetDigitalMode()     do { ANSELAbits.ANSA2 = 0; } while(0)

// get/set IO_RA3 aliases
#define IO_RA3_TRIS                 TRISAbits.TRISA3
#define IO_RA3_PORT                 PORTAbits.RA3
#define IO_RA3_WPU                  WPUAbits.WPUA3
#define IO_RA3_GetValue()           PORTAbits.RA3
#define IO_RA3_SetDigitalInput()    do { TRISAbits.TRISA3 = 1; } while(0)
#define IO_RA3_SetDigitalOutput()   do { TRISAbits.TRISA3 = 0; } while(0)
#define IO_RA3_SetPullup()          do { WPUAbits.WPUA3 = 1; } while(0)
#define IO_RA3_ResetPullup()        do { WPUAbits.WPUA3 = 0; } while(0)

// get/set IO_RA4 aliases
#define IO_RA4_TRIS                 TRISAbits.TRISA4
#define IO_RA4_LAT                  LATAbits.LATA4
#define IO_RA4_PORT                 PORTAbits.RA4
#define IO_RA4_WPU                  WPUAbits.WPUA4
#define IO_RA4_ANS                  ANSELAbits.ANSA4
#define IO_RA4_SetHigh()            do { LATAbits.LATA4 = 1; } while(0)
#define IO_RA4_SetLow()             do { LATAbits.LATA4 = 0; } while(0)
#define IO_RA4_Toggle()             do { LATAbits.LATA4 = ~LATAbits.LATA4; } while(0)
#define IO_RA4_GetValue()           PORTAbits.RA4
#define IO_RA4_SetDigitalInput()    do { TRISAbits.TRISA4 = 1; } while(0)
#define IO_RA4_SetDigitalOutput()   do { TRISAbits.TRISA4 = 0; } while(0)
#define IO_RA4_SetPullup()          do { WPUAbits.WPUA4 = 1; } while(0)
#define IO_RA4_ResetPullup()        do { WPUAbits.WPUA4 = 0; } while(0)
#define IO_RA4_SetAnalogMode()      do { ANSELAbits.ANSA4 = 1; } while(0)
#define IO_RA4_SetDigitalMode()     do { ANSELAbits.ANSA4 = 0; } while(0)

// get/set IO_RA5 aliases
#define IO_RA5_TRIS                 TRISAbits.TRISA5
#define IO_RA5_LAT                  LATAbits.LATA5
#define IO_RA5_PORT                 PORTAbits.RA5
#define IO_RA5_WPU                  WPUAbits.WPUA5
#define IO_RA5_SetHigh()            do { LATAbits.LATA5 = 1; } while(0)
#define IO_RA5_SetLow()             do { LATAbits.LATA5 = 0; } while(0)
#define IO_RA5_Toggle()             do { LATAbits.LATA5 = ~LATAbits.LATA5; } while(0)
#define IO_RA5_GetValue()           PORTAbits.RA5
#define IO_RA5_SetDigitalInput()    do { TRISAbits.TRISA5 = 1; } while(0)
#define IO_RA5_SetDigitalOutput()   do { TRISAbits.TRISA5 = 0; } while(0)
#define IO_RA5_SetPullup()          do { WPUAbits.WPUA5 = 1; } while(0)
#define IO_RA5_ResetPullup()        do { WPUAbits.WPUA5 = 0; } while(0)

typedef void (*interruptHandler)(void);
void (*i2c_driver_busCollisionISR)(void);
void (*i2c_driver_i2cISR)(void);

inline void i2c_driver_close(void)
{
    SSP1CON1bits.SSPEN = 0;
}

/* Interrupt Control */
inline void mssp_enableIRQ(void)
{
    PIE1bits.SSP1IE = 1;
}

inline __bit mssp_IRQisEnabled(void)
{
    return PIE1bits.SSP1IE;
}

inline void mssp_disableIRQ(void)
{
    PIE1bits.SSP1IE = 0;
}

inline void mssp_clearIRQ(void)
{
    PIR1bits.SSP1IF = 0;
}

inline __bit mssp_IRQisSet(void)
{
    return PIR1bits.SSP1IF;
}

inline void mssp_waitForEvent(uint16_t timeout)
{
    uint16_t to = timeout;

    if (PIR1bits.SSP1IF == 0)
    {
        while(to--)
        {
            if (PIR1bits.SSP1IF) break;
            __delay_us(100);
        }
    }
}

__bit i2c_driver_open(void)
{
    if (!SSP1CON1bits.SSPEN)
    {
        SSP1STAT = 0x00;
        SSP1CON1 = 0x28;
        SSP1CON2 = 0x00;
        SSP1ADD = 0x13;
        return true;
    }
    else
        return false;
}

__bit i2c_driver_initSlaveHardware(void)
{
    if(!SSP1CON1bits.SSPEN)
    {
/* NOTE on AHEN:
 * If multiple slaves are to be emulated, then AHEN must be set.  It must be set
 * because the driver needs to selectively ACK/NACK the address depending on its
 * ability to handle the address.
*/

/* NOTE on DHEN:
 * DHEN must be set so that the data is not automatically NACK'ed if it is not read
 * from the SSPBUF.  This driver will ALWAYS read the SSPBUF so that it can pass
 * the value to the appropriate slave handler.  Because the data is ALWAYS read
 * the data will always be ACK'd if DHEN is cleared.  If the slave does not want
 * the data byte from the master then it will return false and a NACK will be returned.
 */

/* NOTE on SEN:
 * SEN will be set enabling clock stretching.  This is because we don't know how
 * long the user will take to process data bytes in their callbacks.  If they are fast,
 * we may not need to stretch the clock.  If they are slow, we need to stretch the clock.
 * If we ALWAYS stretch the clock, we will release the clock when the ISR is complete.
 */

/* NOTE on PCIE:
 * PCIE will be set to enable interrupts on STOP.  This will allow us know when
 * the master is finished
 */
        
/* NOTE on SCIE:
 * SCIE will be set to enable interrupts on START.  This will allow us to detect
 * both a START and a RESTART event and prepare to restart communications.
 */
        SSP1CON1 |= 0x06; //setup I2C Slave (7-bit Addressing)
        SSP1STAT = 0x00;
        SSP1CON2 = 0x00;
        
        SSP1CON1bits.SSPEN = 1;
        return true;
    }
    return false;
}

inline void i2c_driver_resetBus(void)
{
    
}

inline void i2c_driver_start(void)
{
    SSP1CON2bits.SEN = 1;
}

inline void i2c_driver_restart(void)
{
    SSP1CON2bits.RSEN = 1;
}

inline void i2c_driver_stop(void)
{
    SSP1CON2bits.PEN = 1;
}

inline __bit i2c_driver_isNACK(void)
{
    return SSP1CON2bits.ACKSTAT;
}

inline void i2c_driver_startRX(void)
{
    SSP1CON2bits.RCEN = 1;
}

inline char i2c_driver_getRXData(void)
{
    return SSP1BUF;
}

inline void i2c_driver_setAddr(char addr)
{
    SSP1ADD = addr;
}

inline void i2c_driver_setMask(char mask)
{
    SSP1MSK = mask;
}

inline void i2c_driver_TXData(char d)
{
    SSP1BUF = d;
}

inline char i2c_driver_getAddr(void)
{
    return SSP1ADD;
}

inline void i2c_driver_sendACK(void)
{
    SSP1CON2bits.ACKDT = 0;
    SSP1CON2bits.ACKEN = 1; // start the ACK/NACK
}

inline void i2c_driver_sendNACK(void)
{
    SSP1CON2bits.ACKDT = 1;
    SSP1CON2bits.ACKEN = 1; // start the ACK/NACK
}

inline void i2c_driver_releaseClock(void)
{
    SSP1CON1bits.CKP = 1;
}

inline __bit i2c_driver_isBufferFull(void)
{
    return SSP1STATbits.BF;
}

inline __bit i2c_driver_isStart(void)
{
    return SSP1STATbits.S;
}

inline __bit i2c_driver_isAddress(void)
{
    return !SSP1STATbits.D_nA;
}

inline __bit i2c_driver_isStop(void)
{
    return SSP1STATbits.P;
}

inline __bit i2c_driver_isData(void)
{
    return SSP1STATbits.D_nA;
}

inline __bit i2c_driver_isRead(void)
{
    return SSP1STATbits.R_nW;
}

inline __bit i2c_driver_isWriteCollision(void)
{
    return SSP1CON1bits.WCOL;
}

inline __bit i2c_driver_isReceiveOverflow(void)
{
    return SSP1CON1bits.SSPOV;
}

inline void i2c_driver_clearBusCollision(void)
{
    PIR2bits.BCL1IF = 0; // clear the bus collision.
}

inline void i2c_driver_setBusCollisionISR(interruptHandler handler)
{
    i2c_driver_busCollisionISR = handler;
}

inline void i2c_driver_setI2cISR(interruptHandler handler)
{
    i2c_driver_i2cISR = handler;
}

void I2C_Init(uint32_t i2c_clk_freq)
{
	SSP1STAT = 0x00;
	SSP1CON1 = 0x28;
	SSP1CON2 = 0x00;
	SSP1ADD = (_XTAL_FREQ / (4 * i2c_clk_freq)) - 1;  // set I2C clock frequency
}

void I2C_Start(void)
{
	mssp_clearIRQ();
	i2c_driver_start();
	mssp_waitForEvent(500);
}

void I2C_Stop(void)
{
	mssp_clearIRQ();
	i2c_driver_stop();
	mssp_waitForEvent(500);
}

void I2C_Write(uint8_t i2c_data)
{
	mssp_clearIRQ();
	SSP1BUF = i2c_data;
	mssp_waitForEvent(500);
}

uint8_t I2C_Read(uint8_t ack)
{
	uint8_t _data;
	while ((SSPSTAT & 0x04) || (SSPCON2 & 0x1F));  // wait for MSSP module to be free (not busy)
	RCEN = 1;
	while ((SSPSTAT & 0x04) || (SSPCON2 & 0x1F));  // wait for MSSP module to be free (not busy)
	_data = SSPBUF;  // read data from buffer
	while ((SSPSTAT & 0x04) || (SSPCON2 & 0x1F));  // wait for MSSP module to be free (not busy)
	// send acknowledge pulse ? (depends on ack, if 1 send, otherwise don't send)
	ACKDT = !ack;
	ACKEN = 1;
	return _data;  // return data read
}

void write_to_i2c(uint8_t i2c_address, uint8_t command, uint8_t chr)
{ 
	I2C_Start();
	I2C_Write(i2c_address);
	I2C_Write(command);
	I2C_Write(chr);
	I2C_Stop();
}

/**
   @Param
    none
   @Returns
    none
   @Description
    GPIO and peripheral I/O initialization
   @Example
    PIN_MANAGER_Initialize();
 */
void PIN_MANAGER_Initialize(void)
{
    /**
    LATx registers
    */
    LATA = 0x00;

    /**
    TRISx registers
    */
    TRISA = 0b00111111;

    /**
    ANSELx registers
    */
    ANSELA = 0x00;

    // Global WPUEN bit of the OPTION register must be cleared for individual pull-ups to be enabled
    // in our case we want to disable global pull-ups
    OPTION_REGbits.nWPUEN = 0;

    /**
    WPUx registers
    */
    WPUA = 0b00110001;
    // RA0 is button stdby, pullup
    // RA1 is SCL, no pullup
    // RA2 is SDA, no pullup
    // RA3 is ir input, no pullup
    // RA4 is volume+, pullup
    // RA5 is volume-, pullup

    /**
    APFCONx registers
    */
    APFCON = 0x00;    
}

/**
 * @Param
    none
 * @Returns
    none
 * @Description
    This macro will enable global interrupts.
 * @Example
    INTERRUPT_GlobalInterruptEnable();
 */
#define INTERRUPT_GlobalInterruptEnable() (INTCONbits.GIE = 1)

/**
 * @Param
    none
 * @Returns
    none
 * @Description
    This macro will disable global interrupts.
 * @Example
    INTERRUPT_GlobalInterruptDisable();
 */
#define INTERRUPT_GlobalInterruptDisable() (INTCONbits.GIE = 0)
/**
 * @Param
    none
 * @Returns
    none
 * @Description
    This macro will enable peripheral interrupts.
 * @Example
    INTERRUPT_PeripheralInterruptEnable();
 */
#define INTERRUPT_PeripheralInterruptEnable() (INTCONbits.PEIE = 1)

/**
 * @Param
    none
 * @Returns
    none
 * @Description
    This macro will disable peripheral interrupts.
 * @Example
    INTERRUPT_PeripheralInterruptDisable();
 */
#define INTERRUPT_PeripheralInterruptDisable() (INTCONbits.PEIE = 0)

/**
 * @Param
    none
 * @Returns
    none
 * @Description
    Initializes the oscillator to the default states configured in the
 *                  MCC GUI
 * @Example
    OSCILLATOR_Initialize(void);
 */
void OSCILLATOR_Initialize(void)
{
    // SPLLEN enabled; IRCF 32MHz_HF; SCS Clock determined by FOSC<2:0> in Configuration Word 1; 
    OSCCON = 0b11110000;
    // TUN 0; 
    OSCTUNE = 0x00;
    // SBOREN disabled; 
    BORCON = 0x00;
}

/**
 * @Param
    none
 * @Returns
    none
 * @Description
    Initializes the WDT module to the default states configured in the
 *                  MCC GUI
 * @Example
    WDT_Initialize(void);
 */
void WDT_Initialize(void)
{
    // WDTPS 1:65536; SWDTEN OFF; 
    WDTCON = 0x16;
}


/**
 * @Param
    none
 * @Returns
    none
 * @Description
    Initializes the device to the default states configured in the
 *                  MCC GUI
 * @Example
    SYSTEM_Initialize(void);
 */
void SYSTEM_Initialize(void)
{

    PIN_MANAGER_Initialize();
    OSCILLATOR_Initialize();
    WDT_Initialize();
}

#define MCP4725_I2C_ADDRESS	0x60

#define MCP4726_CMD_WRITEDAC            0x40  // Writes data to the DAC
#define MCP4726_CMD_WRITEDAC_PD         0x46  // Writes data to the DAC and enters MCP into power down mode with 500k
#define MCP4726_CMD_WRITEDACEEPROM      0x60  // Writes data to the DAC and the EEPROM (persisting the assigned value after reset)

volatile uint16_t mcp_value = 4095; // default max
volatile uint16_t mcp_value_bak = 0; // backup value, will be changed at standby command

uint16_t MCP4725_read(uint8_t address)
{
	uint8_t buffer[3];
	uint16_t value;
	
	I2C_Start();
	I2C_Write((address << 1) + 1);
	buffer[0] = I2C_Read(1);
	buffer[1] = I2C_Read(1);
	buffer[2] = I2C_Read(0);
	I2C_Stop();
	value = buffer[1];
	value = value << 4;
	value = value + (buffer[2] >> 4);
	return value;
}

void MCP4725_write(uint8_t address, uint16_t val, bool writeEEPROM, bool shutdown)
{
	uint8_t buffer[4];

	buffer[0] = (address << 1);

	if (writeEEPROM)                                // command and config bits  (C2.C1.C0.x.x.PD1.PD0.x)
		buffer[1] = MCP4726_CMD_WRITEDACEEPROM;
    else if (shutdown)
        buffer[1] = MCP4726_CMD_WRITEDAC_PD;
	else
		buffer[1] = MCP4726_CMD_WRITEDAC;

	buffer[2] = (val / 16);                         // Upper data bits          (D11.D10.D9.D8.D7.D6.D5.D4)
	buffer[3] = (val % 16) << 4;                    // Lower data bits          (D3.D2.D1.D0.x.x.x.x)

	I2C_Start();
	I2C_Write(buffer[0]);
	I2C_Write(buffer[1]);
	I2C_Write(buffer[2]);
	I2C_Write(buffer[3]);
	I2C_Stop();
}

#define PCA9536_I2C_ADDRESS 0x41

#define INPUT_PORT_REGISTER 0x00
#define OUTPUT_PORT_REGISTER 0x01
#define POLARITY_INVERSION_REGISTER 0x02
#define CONFIGURATION_REGISTER 0x03

uint8_t PCA9536_Read(uint8_t reg)
{
	uint8_t buffer[1];

	I2C_Start();
	I2C_Write(PCA9536_I2C_ADDRESS << 1);
	I2C_Write(reg);
	I2C_Stop();

    I2C_Start();
	I2C_Write((PCA9536_I2C_ADDRESS << 1) + 1);
	buffer[0] = I2C_Read(0);
	I2C_Stop();

	return (uint8_t)buffer[0];
}

void PCA9536_Write(uint8_t reg, uint8_t data)
{
	I2C_Start();
	I2C_Write(PCA9536_I2C_ADDRESS << 1);
	I2C_Write(reg);
    I2C_Write(data);
	I2C_Stop();
}

/*------------------------------ OLED 128x32 definitions ---------------------------------*/

#define SSD1306_I2C_ADDRESS	     (0x3C << 1)

#define SSD1306_LCDWIDTH             128
#define SSD1306_LCDHEIGHT            32

#define SSD1306_NORMALDISPLAY        0xA6
#define SSD1306_INVERTDISPLAY        0xA7
#define SSD1306_DISPLAYOFF           0xAE
#define SSD1306_DISPLAYON            0xAF
#define SSD1306_SETDISPLAYOFFSET     0xD3
#define SSD1306_SETCOMPINS           0xDA
#define SSD1306_SETVCOMDETECT        0xDB
#define SSD1306_SETDISPLAYCLOCKDIV   0xD5
#define SSD1306_SETPRECHARGE         0xD9
#define SSD1306_SETCONTRAST          0x81
#define SSD1306_DISPLAYALLON_RESUME  0xA4
#define SSD1306_SETMULTIPLEX         0xA8
#define SSD1306_SETLOWCOLUMN         0x00
#define SSD1306_SETHIGHCOLUMN        0x10
#define SSD1306_SETSTARTLINE         0x40
#define SSD1306_MEMORYMODE           0x20
#define SSD1306_COLUMNADDR           0x21
#define SSD1306_PAGEADDR             0x22
#define SSD1306_COMSCANINC           0xC0
#define SSD1306_COMSCANDEC           0xC8
#define SSD1306_SEGREMAP             0xA0
#define SSD1306_CHARGEPUMP           0x8D
#define SSD1306_EXTERNALVCC          0x01
#define SSD1306_SWITCHCAPVCC         0x02

// Scrolling #defines
#define SSD1306_ACTIVATE_SCROLL      0x2F
#define SSD1306_DEACTIVATE_SCROLL                    0x2E
#define SSD1306_SET_VERTICAL_SCROLL_AREA             0xA3
#define SSD1306_RIGHT_HORIZONTAL_SCROLL              0x26
#define SSD1306_LEFT_HORIZONTAL_SCROLL               0x27
#define SSD1306_VERTICAL_AND_RIGHT_HORIZONTAL_SCROLL 0x29
#define SSD1306_VERTICAL_AND_LEFT_HORIZONTAL_SCROLL  0x2A

#define Set_Page_Start_Address_CMD                   0xB0

#define y_max 4

#define CMD	0x00
#define DAT	0x40

volatile bool display_is_on = false;

// Init sequence for Adafruit 128x32 OLED module
uint8_t ssd1306_init_sequence[] = {
	SSD1306_DISPLAYOFF,               // 0xAE
	SSD1306_SETDISPLAYCLOCKDIV,       // 0xD5
	0x80,                             // the suggested ratio 0x80
	SSD1306_SETMULTIPLEX,             // 0xA0
	0x1F,                             // ratio 32
	SSD1306_SETDISPLAYOFFSET,         // 0xD3
	0x00,                             // no offset
	SSD1306_SETSTARTLINE | 0x00,      // line #0
	SSD1306_CHARGEPUMP,               // 0x8D
	0x14,                             // internal vcc
	SSD1306_MEMORYMODE,               // 0x20
	0x02,                             // page mode
	SSD1306_SEGREMAP | 0x1,           // column 127 mapped to SEG0
	SSD1306_COMSCANDEC,               // column scan direction reversed
	SSD1306_SETCOMPINS,               // 0xDA
	0x02,                             // sequential COM pins, disable remap
	SSD1306_SETCONTRAST,              // 0x81
	25,	//0x7F,                       // contrast level 127
	SSD1306_SETPRECHARGE,             // 0xD9
	0xF1,                             // pre-charge period (1, 15)
	SSD1306_SETVCOMDETECT,            // 0xDB
	0x40,                             // vcomh regulator level
	SSD1306_DISPLAYALLON_RESUME,      // 0xA4
	SSD1306_NORMALDISPLAY,            // 0xA6
    SSD1306_DISPLAYOFF                // 0xAE
};

void ssd1306_write(uint8_t addr, uint8_t cmd)
{
	write_to_i2c(SSD1306_I2C_ADDRESS, addr, cmd);
}

void ssd1306_init(void)
{
	uint8_t i;
	
	for (i=0; i < sizeof(ssd1306_init_sequence); ++i)
		ssd1306_write(CMD, ssd1306_init_sequence[i]);
}

#define FONT_WIDTH 11
#define FONT_HEIGHT 16

const uint8_t font_bitmap[] = {
	0x00, 0xF0, 0x08, 0x04, 0x04, 0x04, 0x04, 0x08, 0xF0, 0x00, 0x00, 0x00, 0x0F, 0x10, 0x20, 0x20, 0x20, 0x20, 0x10, 0x0F, 0x00, 0x00, //0
	0x00, 0x00, 0x00, 0x10, 0x10, 0xFC, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x20, 0x3F, 0x20, 0x20, 0x00, 0x00, 0x00, //1
	0x00, 0x00, 0x08, 0x04, 0x04, 0x04, 0x04, 0x88, 0x70, 0x00, 0x00, 0x00, 0x00, 0x30, 0x28, 0x24, 0x22, 0x21, 0x20, 0x20, 0x00, 0x00, //2
	0x00, 0x00, 0x08, 0x04, 0x84, 0x84, 0x84, 0x48, 0x30, 0x00, 0x00, 0x00, 0x00, 0x10, 0x20, 0x20, 0x20, 0x20, 0x11, 0x0E, 0x00, 0x00, //3
	0x00, 0x00, 0x80, 0x40, 0x20, 0x10, 0x08, 0xFC, 0x00, 0x00, 0x00, 0x00, 0x03, 0x02, 0x02, 0x02, 0x02, 0x02, 0x3F, 0x02, 0x00, 0x00, //4
	0x00, 0x00, 0x7C, 0x44, 0x44, 0x44, 0x44, 0x84, 0x04, 0x00, 0x00, 0x00, 0x00, 0x10, 0x20, 0x20, 0x20, 0x20, 0x10, 0x0F, 0x00, 0x00, //5
	0x00, 0xE0, 0x90, 0x48, 0x44, 0x44, 0x44, 0x84, 0x00, 0x00, 0x00, 0x00, 0x0F, 0x10, 0x20, 0x20, 0x20, 0x20, 0x10, 0x0F, 0x00, 0x00, //6
	0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0xC4, 0x34, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x30, 0x0C, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, //7
	0x00, 0x30, 0x48, 0x84, 0x84, 0x84, 0x84, 0x48, 0x30, 0x00, 0x00, 0x00, 0x0E, 0x11, 0x20, 0x20, 0x20, 0x20, 0x11, 0x0E, 0x00, 0x00, //8
	0x00, 0xF0, 0x08, 0x04, 0x04, 0x04, 0x04, 0x08, 0xF0, 0x00, 0x00, 0x00, 0x00, 0x21, 0x22, 0x22, 0x22, 0x12, 0x09, 0x07, 0x00, 0x00, //9
	0x00, 0x0C, 0x70, 0x80, 0x00, 0x00, 0x00, 0x80, 0x70, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x03, 0x1C, 0x20, 0x1C, 0x03, 0x00, 0x00, 0x00, //V
	0xE0, 0x18, 0x08, 0x04, 0x04, 0x04, 0x04, 0x04, 0x08, 0x18, 0xE0, 0x07, 0x18, 0x10, 0x20, 0x20, 0x20, 0x20, 0x20, 0x10, 0x18, 0x07, //O
	0x00, 0x00, 0xFC, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3F, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x00, 0x00, //L
	0xFC, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x00, 0x0F, 0x10, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x10, 0x0F, 0x00, //U
	0xFC, 0x0C, 0x30, 0xC0, 0x00, 0x00, 0x00, 0xC0, 0x30, 0x0C, 0xFC, 0x3F, 0x00, 0x00, 0x00, 0x03, 0x04, 0x03, 0x00, 0x00, 0x00, 0x3F, //M
	0x00, 0xFC, 0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x00, 0x00, 0x00, 0x3F, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x00, 0x00, //E
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x30, 0x00, 0x00, 0x00, 0x00, 0x00 //.
};

const uint16_t font_bitmap_offsets[] = {0, 0x16, 0x2c, 0x42, 0x58, 0x6e, 0x84, 0x9a, 0xb0, 0xc6, 0xdc, 0xf2, 0x108, 0x11e, 0x134, 0x14a, 0x160};

void ssd1306_draw_font(uint8_t x, uint8_t y, uint16_t num)
{
	size_t tt = 0;
	uint8_t x_pos = x;
	uint8_t y_pos = y;

	uint8_t bytes = 0;
	uint16_t ii = 0;
	uint16_t ofset = font_bitmap_offsets[num];

	for(tt=0; tt<22;)
	{
		bytes = font_bitmap[ofset++];

		if (x_pos >= (x + FONT_WIDTH))
		{
			x_pos = x;
			y_pos++;
		}

		ssd1306_write(CMD, Set_Page_Start_Address_CMD + y_pos);        // set row
		ssd1306_write(CMD, SSD1306_SETLOWCOLUMN | (x_pos & 0xf));      // set lower column address
		ssd1306_write(CMD, SSD1306_SETHIGHCOLUMN | (x_pos >> 4));      // set higher column address

		ssd1306_write(DAT, (uint8_t)bytes);
		tt++;
		x_pos++;
	}
}

void draw_volume_text(uint8_t x, uint8_t y)
{
	ssd1306_draw_font(x, y, 10);
	ssd1306_draw_font(x+FONT_WIDTH, y, 11);
	ssd1306_draw_font(x+(2*FONT_WIDTH), y, 12);
    ssd1306_draw_font(x+(3*FONT_WIDTH), y, 13);
    ssd1306_draw_font(x+(4*FONT_WIDTH), y, 14);
    ssd1306_draw_font(x+(5*FONT_WIDTH), y, 15);
}

void draw_volume_value(uint8_t x, uint8_t y)
{
    uint16_t val;
    uint32_t val2;

    if (mcp_value == 4095)
        mcp_value = 4096;

	val2 = 4096 - mcp_value;
	val2 *= 10000;
	val2 /= 128;
	val2 *= 1172;
	val2 /= 10000;
	val2 = 12500 + val2;

    val = (uint16_t)val2;
    
	ssd1306_draw_font(x, y, (val/10000)%10);
	ssd1306_draw_font(x+FONT_WIDTH, y, 16);
	ssd1306_draw_font(x+(2*FONT_WIDTH), y, (val/1000)%10);
	ssd1306_draw_font(x+(3*FONT_WIDTH), y, (val/100)%10);
	ssd1306_draw_font(x+(4*FONT_WIDTH), y, 10);

    if (mcp_value == 4096)
        mcp_value = 4095;
}

void ssd1306_fill(uint8_t bmp_data) 
{                                                     
	uint8_t x_pos;
	uint8_t page;

	for (page=0; page < y_max; ++page)
	{
		ssd1306_write(CMD, (Set_Page_Start_Address_CMD + page));
        ssd1306_write(CMD, SSD1306_SETLOWCOLUMN);
		ssd1306_write(CMD, SSD1306_SETHIGHCOLUMN);

		for (x_pos=0; x_pos < SSD1306_LCDWIDTH; ++x_pos)
		{
			ssd1306_write(DAT, bmp_data);
		}                                                                                  
	}   
}

// 8bit timer
void timer0_initialise(void)
{
    // 50us x (xtail / 4)
    // 0.00005 * 8000000 = 400 ticks ; prescaler=2 = 400/2 = 200 ticks for 50us timer

    // writing to TMR0 resets prescaler, so you need to set prescaler every time!
    TMR0 = 55; // start at 55 ticks (255 - 200)

    OPTION_REGbits.TMR0CS = 0; // Internal instruction cycle clock (FOSC/4)
    OPTION_REGbits.TMR0SE = 0; // low to high
    OPTION_REGbits.PSA = 0; // Prescaler is assigned to the Timer0 module
    OPTION_REGbits.PS2 = 0; // prescaler 2 (PS2:PS1:PS0)
    OPTION_REGbits.PS1 = 0;
    OPTION_REGbits.PS0 = 0;
    // 000 1 : 2
    // 001 1 : 4
    // 010 1 : 8
    // 011 1 : 16
    // 100 1 : 32
    // 101 1 : 64
    // 110 1 : 128
    // 111 1 : 256


    INTCONbits.TMR0IF = 0; // clear interupt first, it will not selfclear!
    INTCONbits.TMR0IE = 1; // enable timer0
}

volatile bool powerbutton = false;
volatile bool volume_plus = false;
volatile bool volume_minus = false;
volatile bool ir_interupt = false;

// RA0 is button stdby, pullup
// RA1 is SCL, no pullup
// RA2 is SDA, no pullup
// RA3 is ir input, no pullup
// RA4 is volume+, pullup
// RA5 is volume-, pullup

void __interrupt() ISR(void)
{
    if (INTCONbits.TMR0IE == 1 && INTCONbits.TMR0IF == 1)
    {
        INTCONbits.TMR0IF = 0; // clear interupt first, it will not selfclear!
        INTCONbits.TMR0IE = 0; // disable timer0
    }

    if (INTCONbits.IOCIE == 1 && INTCONbits.IOCIF == 1)
    {
        if (IOCAFbits.IOCAF0 == 1)
        {
            // interrupt on change for group IOCAF
            IOCAFbits.IOCAF0 = 0;
        
            // interrupt on change for group IOCAN
            IOCANbits.IOCAN0 = 1;
        
            // interrupt on change for group IOCAP
            IOCAPbits.IOCAP0 = 0;

            powerbutton = true;        
        }

        if (IOCAFbits.IOCAF3 == 1)
        {
            // interrupt on change for group IOCAF
            IOCAFbits.IOCAF3 = 0;
        
            // interrupt on change for group IOCAN
            IOCANbits.IOCAN3 = 1;
        
            // interrupt on change for group IOCAP
            IOCAPbits.IOCAP3 = 1;
 
            ir_interupt = true;       
        }

        if (IOCAFbits.IOCAF4 == 1)
        {
            // interrupt on change for group IOCAF
            IOCAFbits.IOCAF4 = 0;
        
            // interrupt on change for group IOCAN
            IOCANbits.IOCAN4 = 1;
        
            // interrupt on change for group IOCAP
            IOCAPbits.IOCAP4 = 0;

            volume_plus = true;        
        }

        if (IOCAFbits.IOCAF5 == 1)
        {
            // interrupt on change for group IOCAF
            IOCAFbits.IOCAF5 = 0;
        
            // interrupt on change for group IOCAN
            IOCANbits.IOCAN5 = 1;
        
            // interrupt on change for group IOCAP
            IOCAPbits.IOCAP5 = 0;

            volume_minus = true;        
        }
    }
}

void process_power(void)
{
    if (display_is_on == false)
    {
        MCP4725_write(MCP4725_I2C_ADDRESS, mcp_value, false, false); // power up MCP
        PCA9536_Write(OUTPUT_PORT_REGISTER, 0b11110000); // power up buck regulator
        ssd1306_write(CMD, SSD1306_DISPLAYON); // power up display
        draw_volume_text(33, 0);
        draw_volume_value(37, 2);
        display_is_on = true;
    }
    else
    {
        PCA9536_Write(OUTPUT_PORT_REGISTER, 0b11110011); // power down buck regulator

        // less writing to MCP eeprom
        if (mcp_value_bak != mcp_value)
        {
            mcp_value_bak = mcp_value;
            MCP4725_write(MCP4725_I2C_ADDRESS, mcp_value, true, false); // store last changed value to the eeprom of the MCP
            __delay_ms(50);
        }

        MCP4725_write(MCP4725_I2C_ADDRESS, mcp_value, false, true); // power down MCP

        ssd1306_write(CMD, SSD1306_DISPLAYOFF); // power down display
        display_is_on = false;
    }
}

void process_volume_plus(void)
{
    if (display_is_on)
    {
        if (mcp_value == 4095)
            mcp_value = 4096;

        if (mcp_value >= 128)
            mcp_value -= 128;

        MCP4725_write(MCP4725_I2C_ADDRESS, mcp_value, false, false);
        draw_volume_value(37, 2);
    }
}

void process_volume_minus(void)
{
    if (display_is_on)
    {
        if (mcp_value == 4095)
            mcp_value = 4096;

        if (mcp_value < 4096)
            mcp_value += 128;

        if (mcp_value >= 4096)
            mcp_value = 4095;

        MCP4725_write(MCP4725_I2C_ADDRESS, mcp_value, false, false);
        draw_volume_value(37, 2);
    }
}

void main(void)
{
    // initialize the device
    SYSTEM_Initialize();

    // When using interrupts, you need to set the Global and Peripheral Interrupt Enable bits
    // Use the following macros to:

    // Enable the Global Interrupts
    INTERRUPT_GlobalInterruptEnable();

    // Enable the Peripheral Interrupts
    INTERRUPT_PeripheralInterruptEnable();

    // Disable the Global Interrupts
    //INTERRUPT_GlobalInterruptDisable();

    // Disable the Peripheral Interrupts
    //INTERRUPT_PeripheralInterruptDisable();

    I2C_Init(400000);    // initialize I2C bus with clock frequency of 400kHz

	__delay_ms(50);

    PCA9536_Write(CONFIGURATION_REGISTER, 0b11110000);
    PCA9536_Write(OUTPUT_PORT_REGISTER, 0b11110011); // power down buck regulator
    mcp_value = MCP4725_read(MCP4725_I2C_ADDRESS);
    MCP4725_write(MCP4725_I2C_ADDRESS, mcp_value, false, true); // power down MCP

    __delay_ms(100);
    ssd1306_init();
    __delay_ms(100);
    ssd1306_fill(0);

    uint32_t usecconds = 0;
  	bool start_bit = false;
  	bool data_bit = false;
  	bool repeating = false;
  	uint8_t count = 0;
  	uint32_t ir_code = 0;
  	uint32_t i = 0;
  	//uint32_t address = 0;
  	uint16_t command = 0;

    IOCAFbits.IOCAF0 = 0;
    IOCANbits.IOCAN0 = 1;
    IOCAPbits.IOCAP0 = 0;

    IOCAFbits.IOCAF3 = 0;
    IOCANbits.IOCAN3 = 1;
    IOCAPbits.IOCAP3 = 1;

    IOCAFbits.IOCAF4 = 0;
    IOCANbits.IOCAN4 = 1;
    IOCAPbits.IOCAP4 = 0;

    IOCAFbits.IOCAF5 = 0;
    IOCANbits.IOCAN5 = 1;
    IOCAPbits.IOCAP5 = 0;

    INTCONbits.IOCIE = 1; // Enable IOCI interrupt

	__delay_ms(50);

    while(1)
    {
        if (ir_interupt)
        {
            usecconds = 0;

            switch(IO_RA3_GetValue())
            {
    			case LOW:

                    while(IO_RA3_GetValue() == LOW)
                    {
                        if (INTCONbits.TMR0IE == 0)
                        {
                            timer0_initialise();
                            usecconds += 50;
                        }
                    }

    				if (usecconds > 100000)
    				{
    					count = 0;
    					start_bit = false;
    				}
    				else if (usecconds > 2100 && usecconds < 2400 && !repeating && start_bit && count == 0)
    				{
    					repeating = true;
    				}
    				else if (usecconds > 4350 && usecconds < 4650 && !repeating && start_bit && count == 0)
    				{
    					count += 1;
    					i = 0;
    					ir_code = 0;
    					//address = 0;
    					command = 0;
    					data_bit = false;
    					repeating = false;
    				}
    				else if (usecconds > 400 && usecconds < 700 && !repeating && start_bit && count && data_bit)
    				{
    					ir_code &= ~(1ul << (31 - i));
    					i += 1;
    					count += 1;
    				}
    				else if (usecconds > 1500 && usecconds < 1800 && !repeating && start_bit && count && data_bit)
    				{
    					ir_code |= 1ul << (31 - i);
    					i += 1;
    					count += 1;
    				}
    				else if (count == 33)
    				{
    					data_bit = false;
    					count += 1;
    				}

    				break;

    			case HIGH:

                    while(IO_RA3_GetValue() == HIGH)
                    {
                        if (INTCONbits.TMR0IE == 0)
                        {
                            timer0_initialise();
                            usecconds += 50;
                        }
                    }

    				if (usecconds < 8800 && !start_bit && !repeating)
    				{
    					count = 0;
    				}
    				else if (usecconds > 8800 && usecconds < 9100 && !start_bit)
    				{
    					start_bit = true;
    					data_bit = false;
    					count = 0;
    					i = 0;
    					ir_code = 0;
    					repeating = false;
    				}
    				else if (usecconds > 400 && usecconds < 600 && !repeating && start_bit && count && !data_bit)
    				{
    					data_bit = true;
    					count += 1;
    				}
    				else if (usecconds > 0 && usecconds < 600 && !repeating && start_bit && count == 34)
    				{
    					start_bit = false;
    					data_bit = false;
    					repeating = false;
    					count = 0;
    					i = 0;
    				}

    				break;

    			default:
    				break;
    		}

    		if (count == 34)
    		{
    			//address = ir_code >> 16;
    			command = ir_code >> 8;

rep:
                // power command = 38920
                // input command = 39064
                // menu command  = 38936
                // up command    = 39144
                // down command  = 39016
                // left command  = 39154
                // right command = 39050
                // ok command    = 38922
                // play command  = 39096
                // vol- command  = 38968
                // vol+ command  = 39010
                // mute command  = 39032

    			switch(command)
    			{
    				case 38920:
    					// POWER
    					if (!repeating)
    					{
                            process_power();
    					}
    					break;

    				case 39154:
    					// LEFT
                        process_volume_minus();
    					break;

    				case 39050:
    					// RIGHT
                        process_volume_plus();
    					break;

    				default:
    					break;
    			}

    			if (repeating)
    			{
    				repeating = false;
    			}
    		}

    		if (repeating)
    		{
    			goto rep;
    		}

            ir_interupt = false;
        }

        if (powerbutton)
        {
                process_power();
                powerbutton = false;        
        }

        if (volume_plus)
        {
                process_volume_plus();
                volume_plus = false;        
        }

        if (volume_minus)
        {
                process_volume_minus();
                volume_minus = false;        
        }

        // put device to deep sleep mode
        SLEEP();
    }
}

